#include <gst/gst.h>
#include <glib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <sys/time.h>
#include "gstnvdsmeta.h"
#include "nvbufsurface.h"
#include "nvds_obj_encode.h"
#ifndef PLATFORM_TEGRA
#include "gst-nvmessage.h"
#endif

/* gst to nvds transform function set by user. "data" holds a pointer to
 * NvDsUserMeta */
static gpointer nvdsbatch_gst_to_nvds_meta_transform_func(
    gpointer data, gpointer user_data) {
    NvDsUserMeta *user_meta = (NvDsUserMeta *)data;
    NvDsBatchMeta *src_nvdsbatch_meta =
        (NvDsBatchMeta *)user_meta->user_meta_data;
    NvDsBatchMeta *dst_nvdsbatch_meta =
        (NvDsBatchMeta *)nvds_batch_meta_copy_func(src_nvdsbatch_meta, NULL);
    return (gpointer)dst_nvdsbatch_meta;
}

/* release function set by user to release gst to nvds transformed metadata.
 * "data" holds a pointer to NvDsUserMeta */
static void nvdsbatch_gst_nvds_meta_release_func(
    gpointer data, gpointer user_data) {
    NvDsUserMeta *user_meta = (NvDsUserMeta *)data;
    NvDsBatchMeta *nvdsbatch_meta = (NvDsBatchMeta *)user_meta->user_meta_data;
    nvds_batch_meta_release_func(nvdsbatch_meta, NULL);
}

typedef struct _ImageFrame {
    GstBuffer *gst_buffer;
    uint batch_id;
    uint width;
    uint height;
} ImageFrame;

typedef struct _JpegEncOut {
    unsigned char *p_dst_data;
    unsigned int data_len;
} JpegEncOut;

bool jpeg_encode(const ImageFrame *frame, int quality, JpegEncOut *enc_out) {
    GstBuffer *buf = NULL;
    GstMapInfo map_info;
    NvBufSurface *nvsurface = NULL;
    NvDsObjEncCtxHandle ctx;
    NvDsFrameMeta *pnv_frame_meta = NULL;
    NvDsMetaList *l_obj_meta = NULL;
    NvDsBatchMeta *batch_meta = NULL;
    NvDsObjectMeta *p_obj_meta = NULL;
    NvDsObjectMeta *p_cur_obj_meta = NULL;
    uint batch_id = 0;

    if (NULL == frame || NULL == frame->gst_buffer || 0 == frame->width || 0 == frame->height ||
        NULL == enc_out) {
        printf("jpeg_encode input param error");
        return false;
    }
    buf = (GstBuffer *)frame->gst_buffer;
    batch_id = frame->batch_id;
    batch_meta = gst_buffer_get_nvds_batch_meta(buf);
    if (!gst_buffer_map(buf, &map_info, GST_MAP_READ)) {
        printf("input buffer mapinfo failed");
        return false;
    }

    nvsurface = (NvBufSurface *)map_info.data;
    if (NULL == nvsurface) {
        printf("nvsurface error");
        gst_buffer_unmap(buf, &map_info);
        return false;
    }

    pnv_frame_meta = (NvDsFrameMeta *)nvds_get_nth_frame_meta(
        batch_meta->frame_meta_list, batch_id);
    if (NULL == pnv_frame_meta) {
        printf("nvds_obj_enc_create_context failed");
        gst_buffer_unmap(buf, &map_info);
        return false;
    }
    ctx = nvds_obj_enc_create_context();
    if (NULL == ctx) {
        printf("nvds_obj_enc_create_context failed");
        gst_buffer_unmap(buf, &map_info);
        return false;
    }
    p_obj_meta = nvds_acquire_obj_meta_from_pool(batch_meta);
    nvds_add_obj_meta_to_frame(pnv_frame_meta, p_obj_meta, NULL);
    p_obj_meta->unique_component_id = 1;
    p_obj_meta->object_id = -1;
    p_obj_meta->confidence = -0.1;
    p_obj_meta->base_meta.meta_type = NVDS_OBJ_META;
    p_obj_meta->base_meta.batch_meta = batch_meta;
    sprintf(p_obj_meta->obj_label, "Person");
    p_obj_meta->rect_params.left = 0;
    p_obj_meta->rect_params.top = 0;
    p_obj_meta->rect_params.width = frame->width & (~1);
    p_obj_meta->rect_params.height = frame->height & (~1);
    NvDsObjEncUsrArgs userData = {0};
    userData.saveImg = false;
    userData.attachUsrMeta = true;
    userData.objNum = 1;
    nvds_obj_enc_process(ctx, &userData, nvsurface, p_obj_meta, pnv_frame_meta);
    nvds_obj_enc_finish(ctx);
    // get the jpeg data
    // clang-format off
    for (l_obj_meta = pnv_frame_meta->obj_meta_list; l_obj_meta != NULL;
        l_obj_meta = l_obj_meta->next) {
        // clang-format on
        p_cur_obj_meta = (NvDsObjectMeta *)(l_obj_meta->data);
        NvDsUserMetaList *usrMetaList = p_cur_obj_meta->obj_user_meta_list;
        while (usrMetaList != NULL) {
            NvDsUserMeta *userMetaData = (NvDsUserMeta *)usrMetaList->data;
            if (userMetaData->base_meta.meta_type == NVDS_CROP_IMAGE_META) {
                NvDsObjEncOutParams *enc_jpeg_image =
                    (NvDsObjEncOutParams *)userMetaData->user_meta_data;
                uint8_t *p_dst_data = (uint8_t *)g_malloc(
                    enc_jpeg_image->outLen * sizeof(uint8_t));
                if (NULL == p_dst_data) {
                    printf("g_malloc failed\n");
                    nvds_obj_enc_destroy_context(ctx);
                    gst_buffer_unmap(buf, &map_info);
                    return false;
                }
                memcpy(p_dst_data, enc_jpeg_image->outBuffer,
                    enc_jpeg_image->outLen * sizeof(uint8_t));
                enc_out->p_dst_data = p_dst_data;
                enc_out->data_len = enc_jpeg_image->outLen;
                usrMetaList = NULL;
            } else {
                usrMetaList = usrMetaList->next;
            }
        }
    }
    nvds_obj_enc_destroy_context(ctx);
    gst_buffer_unmap(buf, &map_info);
    return true;
}

int main() {
    GstBuffer *gstbuf = NULL;
    NvBufSurfaceCreateParams create_params;
    NvBufSurface *inter_buf = NULL;
    NvBufSurface *pNewSurface = NULL;
    NvDsBatchMeta *outbatchmeta = NULL;
    NvDsFrameMeta *frame_meta_out = NULL;
    NvDsMeta *meta = NULL;
    GstMapInfo mem_info;
    uint32_t width = 1920;
    uint32_t height = 1080;
    struct timeval tv;
    int num = 1000000;
    static uint32_t frame_num = 0;
    bool ret = false;

    gst_init(NULL, NULL);

    for(int i = 0; i < num; ++i) {
        create_params.gpuId = 0;
        create_params.width = width;
        create_params.height = height;
        create_params.size = 0;
        create_params.isContiguous = true;
        create_params.colorFormat = NVBUF_COLOR_FORMAT_NV12;
        create_params.layout = NVBUF_LAYOUT_PITCH;
        create_params.memType = NVBUF_MEM_DEFAULT;
        if (0 !=
            NvBufSurfaceCreate(&inter_buf, 1, &create_params)) {
            printf("NvBufSurfaceCreate failed\n");
            return -1;
        }
        NvBufSurfaceMemSet(inter_buf, -1, -1, 0);
        frame_num++;
        inter_buf->numFilled = 1;
        inter_buf->batchSize = 1;

        gstbuf = gst_buffer_new_allocate(NULL, sizeof(NvBufSurface), NULL);
        gst_buffer_map(gstbuf, &mem_info, GST_MAP_WRITE);
        pNewSurface = (NvBufSurface *)mem_info.data;
        *pNewSurface = *inter_buf;
        pNewSurface->batchSize = 1;
        pNewSurface->numFilled = 1;
        pNewSurface->surfaceList = &inter_buf->surfaceList[0];
        gst_buffer_unmap(gstbuf, &mem_info);
        
        // add batch meta
        outbatchmeta = nvds_create_batch_meta(1);

        frame_meta_out = nvds_acquire_frame_meta_from_pool(outbatchmeta);
        frame_meta_out->frame_num = frame_num++;
        gettimeofday(&tv, NULL);
        frame_meta_out->ntp_timestamp = tv.tv_sec * 1000000000ll + tv.tv_usec * 1000ll;
        frame_meta_out->source_frame_width = width;
        frame_meta_out->source_frame_height = height;
        frame_meta_out->pad_index = 0;
        frame_meta_out->batch_id = 0;
        frame_meta_out->num_surfaces_per_frame = 1;
        nvds_add_frame_meta_to_batch(outbatchmeta, frame_meta_out);
        outbatchmeta->max_frames_in_batch = 1;
        outbatchmeta->num_frames_in_batch = 1;
        meta = gst_buffer_add_nvds_meta(gstbuf, outbatchmeta, NULL,
            nvds_batch_meta_copy_func, nvds_batch_meta_release_func);
        meta->meta_type = NVDS_BATCH_GST_META;
        meta->gst_to_nvds_meta_transform_func =
            nvdsbatch_gst_to_nvds_meta_transform_func;
        meta->gst_to_nvds_meta_release_func =
            nvdsbatch_gst_nvds_meta_release_func;
        if (i % 10000 == 0) {
            printf("nvds_obj_enc_test,i=%d\n", i);
        }

        JpegEncOut jpeg_out = {NULL, 0};
        ImageFrame frame;

        // use nvds_obj_enc_process interface to do batch jpeg encode
        frame.gst_buffer = gstbuf;
        frame.batch_id = 0;
        frame.width = width;
        frame.height = height;
        ret = jpeg_encode(&frame, 75, &jpeg_out);
        if (true == ret) {
            if (jpeg_out.p_dst_data) {
                g_free(jpeg_out.p_dst_data);
            }
        }

        if (inter_buf) {
            NvBufSurfaceDestroy(inter_buf);
        }
	
        gst_buffer_unref(gstbuf);
    }
}
